from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init

try:
    import ydata_profiling
except ModuleNotFoundError:
    pass


class ProfileReport(DataSpecWrapper[ydata_profiling.ProfileReport]):
    @sarus_init("pandas_profiling.PD_PROFILE_REPORT")
    def __init__(self, df=None, **kwargs) -> None:
        ...


register_ops()



"""Sarus Imblearn package documentation."""
# flake8: noqa
try:
    from imblearn import *
except ModuleNotFoundError:
    pass

try:
    import sarus.imblearn.over_sampling as over_sampling
    import sarus.imblearn.pipeline as pipeline
    import sarus.imblearn.under_sampling as under_sampling
except NameError:
    pass



from __future__ import annotations

import logging
from typing import Any, Callable, Dict, Optional, Union

import numpy as np
import pandas as pd

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import sarus_init, sarus_method

try:
    import shap
except ModuleNotFoundError:
    pass

logger = logging.getLogger(__name__)


class Masker(DataSpecWrapper[shap.maskers.Masker]):
    @sarus_init("shap.SHAP_MASKER")
    def __init__(self, **kwargs):
        ...


class Independent(DataSpecWrapper[shap.maskers.Independent]):
    @sarus_init("shap.SHAP_INDEPENDENT")
    def __init__(
        self,
        data: Union[np.ndarray, pd.DataFrame],
        max_samples: Optional[int] = 100,
    ):
        ...

    @sarus_method("shap.SHAP_INVARIANTS")
    def invariants(self, x: Union[np.ndarray, pd.DataFrame]) -> Any:
        ...


class Partition(DataSpecWrapper[shap.maskers.Partition]):
    @sarus_init("shap.SHAP_MASKER_PARTITION")
    def __init__(
        self,
        data: Union[np.ndarray, pd.DataFrame],
        max_samples: int = 100,
        clustering: str = "correlation",
    ):
        ...

    @sarus_method("shap.SHAP_PARTITION_INVARIANTS")
    def invariants(self, x: np.ndarray):
        ...


class Impute(DataSpecWrapper[shap.maskers.Impute]):
    @sarus_init("shap.SHAP_IMPUTE")
    def __init__(
        self,
        data: Union[np.ndarray, pd.DataFrame, Dict[str, np.ndarray]],
        method: str = "linear",
    ):
        ...


class Fixed(DataSpecWrapper[shap.maskers.Fixed]):
    @sarus_init("shap.SHAP_FIXED")
    def __init__(self):
        ...

    @sarus_method("shap.SHAP_FIXED_MASK_SHAPES")
    def invariants(
        self, X: Union[np.ndarray, pd.DataFrame, Dict[str, np.ndarray]]
    ):
        ...


class Composite(DataSpecWrapper[shap.maskers.Composite]):
    @sarus_init("shap.SHAP_COMPOSITE")
    def __init__(self, mask1: Any, mask2: Any = None, mask3: Any = None):
        ...


class FixedComposite(DataSpecWrapper[shap.maskers.FixedComposite]):
    @sarus_init("shap.SHAP_FIXED_COMPOSITE")
    def __init__(self, masker: Any):
        ...


class SHAP_OutputComposite(DataSpecWrapper[shap.maskers.OutputComposite]):
    @sarus_init("shap.SHAP_OUTPUT_COMPOSITE")
    def __init__(self, masker: shap.maskers.Masker, model: shap.models.Model):
        ...


# SHAP Text Masker
class TextMasker:
    @sarus_init("shap.SHAP_TEXT_MASKER")
    def __init__(
        self,
        tokenizer: Optional[Union[Callable, None]] = None,
        mask_token: Optional[Union[str, int, None]] = None,
        collapse_mask_token: Optional[Union[bool, str]] = "auto",
        output_type: Optional[str] = "string",
    ):
        ...

    @sarus_method("shap.SHAP_CLUSTERING")
    def clustering(self, s: str):
        ...

    @sarus_method("shap.SHAP_DATA_TEXT_TRANSFORM")
    def data_transform(self, s: str):
        ...

    @sarus_method("shap.SHAP_FEATURE_NAMES")
    def feature_names(self, s: str):
        ...

    @sarus_method("shap.SHAP_TEXT_INVARIANTS")
    def invariants(self, s: str):
        ...

    @sarus_method("shap.SHAP_MASK_TEXT_SHAPES")
    def mask_shapes(self, s: str):
        ...

    @sarus_method("shap.SHAP_SHAPE")
    def shape(self, s: str):
        ...

    @sarus_method("shap.SHAP_TOKEN_SEGMENTS")
    def token_segments(self, s: str):
        ...


# SHAP Image Masker
class ImageMasker:
    @sarus_init("shap.SHAP_IMAGE_MASKER")
    def __init__(
        self,
        mask_value: Union[np.array, str] = None,
        shape: Optional[tuple] = None,
    ):
        ...

    @sarus_method("shap.SHAP_BUILD_PARTITION_TREE")
    def build_partition_tree(self):
        ...

    @sarus_method("shap.SHAP_INPAINT")
    def inpaint(
        self,
        x: Union[pd.Series, pd.DataFrame, np.array],
        mask: Union[pd.Series, pd.DataFrame, np.array],
        method: str,
    ):
        ...



from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init

try:
    import sklearn.model_selection as sk_model_selection
    from sklearn.model_selection import *  # noqa: F401
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class TimeSeriesSplit(DataSpecWrapper[sk_model_selection.TimeSeriesSplit]):
    @sarus_init("sklearn.SK_TIME_SERIES_SPLIT")
    def __init__(
        self,
        *,
        n_splits=5,
        n_repeats=10,
        random_state=None,
        _dataspec=None,
    ):
        ...


class RepeatedStratifiedKFold(
    DataSpecWrapper[sk_model_selection.RepeatedStratifiedKFold]
):
    @sarus_init("sklearn.SK_REPEATED_STRATIFIED_KFOLD")
    def __init__(
        self,
        *,
        n_splits=5,
        n_repeats=10,
        random_state=None,
        _dataspec=None,
    ):
        ...


class KFold(DataSpecWrapper[sk_model_selection.KFold]):
    @sarus_init("sklearn.SK_KFOLD")
    def __init__(self, n_splits=5, *, shuffle=False, random_state=None):
        ...


register_ops()



from __future__ import annotations

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init, sarus_method

try:
    import sklearn.linear_model as linear_model
    from sklearn.linear_model import *  # noqa: F401
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class LinearRegression(DataSpecWrapper[linear_model.LinearRegression]):
    @sarus_init("sklearn.SK_LINEAR_REGRESSION")
    def __init__(self, steps, *, memory=None, verbose=False):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class LogisticRegression(DataSpecWrapper[linear_model.LogisticRegression]):
    @sarus_init("sklearn.SK_LOGISTIC_REGRESSION")
    def __init__(
        self,
        penalty="l2",
        dual=False,
        tol=0.0001,
        C=1.0,
        fit_intercept=True,
        intercept_scaling=1,
        class_weight=None,
        random_state=None,
        solver="lbfgs",
        max_iter=100,
        multi_class="auto",
        verbose=0,
        warm_start=False,
        n_jobs=None,
        l1_ratio=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


register_ops()



"""SDK classes and functions."""
# flake8: noqa
import warnings

VERSION = "0.9.3rc0"

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from sarus import (
        imblearn,
        numpy,
        pandas,
        pandas_profiling,
        plotly,
        scipy,
        shap,
        sklearn,
        skopt,
        std,
        xgboost,
        optbinning,
    )

    from .sarus import Client, Dataset
    from .utils import (
        eval,
        eval_perturbation,
        eval_policy,
        floating,
        integer,
        length,
    )


__all__ = [
    "Dataset",
    "Client",
    "length",
    "eval",
    "eval_perturbation",
    "eval_policy",
    "config",
    "floating",
    "integer",
]



"""Sarus optbinning package documentation."""
# flake8: noqa
from sarus.utils import register_ops

try:
    from optbinning import *
    from .binning import *

except ModuleNotFoundError:
    pass


register_ops()



import logging
import typing as t

import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st

from sarus.typing import Client

logger = logging.getLogger()


def dot(client: Client, uuid: str) -> t.Any:
    referrable = client.context().storage().referrable(uuid)
    if referrable is None:
        logger.error(f"{uuid} is not found in the Client storage.")

    if referrable.prototype() not in [sp.Dataset, sp.Scalar]:
        logger.error(f"{uuid} is not a Dataspec: {referrable.prototype()}")

    dataspec = t.cast(st.DataSpec, referrable)

    try:
        from graphviz import Source
    except:  # noqa: E722
        logger.warning("Graphviz not installed. Cannot plot dot graph.")
        return

    return Source(dataspec.dot())



from __future__ import annotations

import numpy as np

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init, sarus_method

try:
    from sklearn import cluster
    from sklearn.cluster import *  # noqa: F401
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class AffinityPropagation(DataSpecWrapper[cluster.AffinityPropagation]):
    @sarus_init("sklearn.SK_AFFINITY_PROPAGATION")
    def __init__(
        self,
        *,
        damping=0.5,
        max_iter=200,
        convergence_iter=15,
        copy=True,
        preference=None,
        affinity="euclidean",
        verbose=False,
        random_state="warn",
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class AgglomerativeClustering(
    DataSpecWrapper[cluster.AgglomerativeClustering]
):
    @sarus_init("sklearn.SK_AGGLOMERATIVE_CLUSTERING")
    def __init__(
        self,
        n_clusters=2,
        *,
        affinity="euclidean",
        memory=None,
        connectivity=None,
        compute_full_tree="auto",
        linkage="ward",
        distance_threshold=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class Birch(DataSpecWrapper[cluster.Birch]):
    @sarus_init("sklearn.SK_BIRCH")
    def __init__(
        self,
        *,
        threshold=0.5,
        branching_factor=50,
        n_clusters=3,
        compute_labels=True,
        copy=True,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class DBSCAN(DataSpecWrapper[cluster.DBSCAN]):
    @sarus_init("sklearn.SK_DBSCAN")
    def __init__(
        self,
        eps=0.5,
        *,
        min_samples=5,
        metric="euclidean",
        metric_params=None,
        algorithm="auto",
        leaf_size=30,
        p=None,
        n_jobs=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class FeatureAgglomeration(DataSpecWrapper[cluster.FeatureAgglomeration]):
    @sarus_init("sklearn.SK_AFFINITY_PROPAGATION")
    def __init__(
        self,
        n_clusters=2,
        *,
        affinity="euclidean",
        memory=None,
        connectivity=None,
        compute_full_tree="auto",
        linkage="ward",
        ooling_func=np.mean,
        distance_threshold=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class KMeans(DataSpecWrapper[cluster.KMeans]):
    @sarus_init("sklearn.SK_KMEANS")
    def __init__(
        self,
        n_clusters=8,
        *,
        init="k-means++",
        n_init=10,
        max_iter=300,
        tol=0.0001,
        precompute_distances="deprecated",
        verbose=0,
        random_state=None,
        copy_x=True,
        n_jobs="deprecated",
        algorithm="auto",
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class MiniBatchKMeans(DataSpecWrapper[cluster.MiniBatchKMeans]):
    @sarus_init("sklearn.SK_MINIBATCH_KMEANS")
    def __init__(
        self,
        n_clusters=8,
        *,
        init="k-means++",
        max_iter=100,
        batch_size=100,
        verbose=0,
        compute_labels=True,
        random_state=None,
        tol=0.0,
        max_no_improvement=10,
        init_size=None,
        n_init=3,
        reassignment_ratio=0.01,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class MeanShift(DataSpecWrapper[cluster.MeanShift]):
    @sarus_init("sklearn.SK_MEAN_SHIFT")
    def __init__(
        self,
        *,
        bandwidth=None,
        seeds=None,
        bin_seeding=False,
        min_bin_freq=1,
        cluster_all=True,
        n_jobs=None,
        max_iter=300,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class OPTICS(DataSpecWrapper[cluster.OPTICS]):
    @sarus_init("sklearn.SK_OPTICS")
    def __init__(
        self,
        *,
        min_samples=5,
        max_eps=np.inf,
        metric="minkowski",
        p=2,
        metric_params=None,
        cluster_method="xi",
        eps=None,
        xi=0.05,
        predecessor_correction=True,
        min_cluster_size=None,
        algorithm="auto",
        leaf_size=30,
        n_jobs=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class SpectralClustering(DataSpecWrapper[cluster.SpectralClustering]):
    @sarus_init("sklearn.SK_SPECTRAL_CLUSTERING")
    def __init__(
        self,
        n_clusters=8,
        *,
        eigen_solver=None,
        n_components=None,
        random_state=None,
        n_init=10,
        gamma=1.0,
        affinity="rbf",
        n_neighbors=10,
        eigen_tol=0.0,
        assign_labels="kmeans",
        degree=3,
        coef0=1,
        kernel_params=None,
        n_jobs=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class SpectralBiclustering(DataSpecWrapper[cluster.SpectralBiclustering]):
    @sarus_init("sklearn.SK_SPECTRAL_BICLUSTERING")
    def __init__(
        self,
        n_clusters=3,
        *,
        method="bistochastic",
        n_components=6,
        n_best=3,
        svd_method="randomized",
        n_svd_vecs=None,
        mini_batch=False,
        init="k-means++",
        n_init=10,
        n_jobs="deprecated",
        random_state=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class SpectralCoclustering(DataSpecWrapper[cluster.SpectralCoclustering]):
    @sarus_init("sklearn.SK_AFFINITY_PROPAGATION")
    def __init__(
        self,
        n_clusters=3,
        *,
        svd_method="randomized",
        n_svd_vecs=None,
        mini_batch=False,
        init="k-means++",
        n_init=10,
        n_jobs="deprecated",
        random_state=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


register_ops()



import typing as t

import sarus_data_spec.status as stt
from sarus_data_spec import typing as st
from sarus_data_spec.constants import SCALAR_TASK

from sarus.manager.base_remote import RemoteComputation

from .dataspec_api import dataspec_status, launch_dataspec, scalar_result


class ValueComputationOnServer(RemoteComputation[t.Any]):
    """ValueComputation on the Sarus server through the REST API."""

    task_name = SCALAR_TASK

    def launch_task(self, dataspec: st.DataSpec) -> None:
        """Launch the computation of a Scalar on the server."""
        status = self.status(dataspec=dataspec)
        if status is None:
            launch_dataspec(self.computing_manager().client(), dataspec.uuid())

    async def result_from_stage_properties(
        self,
        dataspec: st.DataSpec,
        properties: t.Mapping[str, str],
        **kwargs: t.Any,
    ) -> t.Any:
        scalar = t.cast(st.Scalar, dataspec)

        # We do not download weights on sarus.eval for LLM
        if scalar.is_pretrained_model() or scalar.is_fitted_model():
            return None

        return scalar_result(
            self.computing_manager().client(), dataspec.uuid()
        )

    def status(self, dataspec: st.DataSpec) -> t.Optional[st.Status]:
        """Retreive the status from the server.

        In this case, the computation retrieves the status from the server
        via an API call, a local status is created but not stored in order
        to use it just for the pollin
        """
        status_proto = dataspec_status(
            client=self.computing_manager().client(),
            uuid=dataspec.uuid(),
            task_names=[self.task_name],
        )
        if status_proto is None:
            return None
        else:
            return stt.Status(protobuf=status_proto, store=False)



import optbinning

from sarus.dataspec_wrapper import (
    DataSpecWrapper,
)
from sarus.utils import sarus_method, sarus_init, sarus_property


class OptimalBinning(DataSpecWrapper[optbinning.OptimalBinning]):
    @sarus_init("optbinning.OPTIMAL_BINNING")
    def __init__(
        self,
        name="",
        dtype="numerical",
        prebinning_method="cart",
        solver="cp",
        divergence="iv",
        max_n_prebins=20,
        min_prebin_size=0.05,
        min_n_bins=None,
        max_n_bins=None,
        min_bin_size=None,
        max_bin_size=None,
        min_bin_n_nonevent=None,
        max_bin_n_nonevent=None,
        min_bin_n_event=None,
        max_bin_n_event=None,
        monotonic_trend="auto",
        min_event_rate_diff=0,
        max_pvalue=None,
        max_pvalue_policy="consecutive",
        gamma=0,
        outlier_detector=None,
        outlier_params=None,
        class_weight=None,
        cat_cutoff=None,
        cat_unknown=None,
        user_splits=None,
        user_splits_fixed=None,
        special_codes=None,
        split_digits=None,
        mip_solver="bop",
        time_limit=100,
        verbose=False,
    ):
        ...

    @sarus_method("optbinning.OPTIMAL_BINNING_FIT", inplace=True)
    def fit(self, x, y, sample_weight=None, check_input=False):
        ...

    @sarus_method("optbinning.OPTIMAL_BINNING_FIT_TRANSFORM")
    def fit_transform(
        self,
        x,
        y,
        sample_weight=None,
        metric="woe",
        metric_special=0,
        metric_missing=0,
        show_digits=2,
        check_input=False,
    ):
        ...

    @sarus_method("optbinning.OPTIMAL_BINNING_TRANSFORM")
    def transform(
        self,
        x,
        y,
        sample_weight=None,
        metric="woe",
        metric_special=0,
        metric_missing=0,
        show_digits=2,
        check_input=False,
    ):
        ...

    @sarus_property("optbinning.OPTIMAL_BINNING_BINNING_TABLE")
    def binning_table(self):
        ...


class BinningProcess(DataSpecWrapper[optbinning.BinningProcess]):
    @sarus_init("optbinning.BINNING_PROCESS")
    def __init__(
        self,
        variable_names,
        max_n_prebins=20,
        min_prebin_size=0.05,
        min_n_bins=None,
        max_n_bins=None,
        min_bin_size=None,
        max_bin_size=None,
        max_pvalue=None,
        max_pvalue_policy="consecutive",
        selection_criteria=None,
        fixed_variables=None,
        categorical_variables=None,
        special_codes=None,
        split_digits=None,
        binning_fit_params=None,
        binning_transform_params=None,
        n_jobs=None,
        verbose=False,
    ):
        ...

    @sarus_method("optbinning.BINNING_PROCESS_FIT", inplace=True)
    def fit(self, X, y, sample_weight=None, check_input=False):
        ...

    @sarus_method("optbinning.BINNING_PROCESS_FIT_TRANSFORM")
    def fit_transform(
        self,
        X,
        y,
        sample_weight=None,
        metric=None,
        metric_special=0,
        metric_missing=0,
        show_digits=2,
        check_input=False,
    ):
        ...

    @sarus_method("optbinning.BINNING_PROCESS_TRANSFORM")
    def transform(
        self,
        X,
        metric=None,
        metric_special=0,
        metric_missing=0,
        show_digits=2,
        check_input=False,
    ):
        ...


class BinningTable(
    DataSpecWrapper[optbinning.binning.binning_statistics.BinningTable]
):
    @sarus_init("optbinning.BINNING_TABLE")
    def __init__(
        self,
        name,
        dtype,
        special_codes,
        splits,
        n_nonevent,
        n_event,
        min_x=None,
        max_x=None,
        categories=None,
        cat_others=None,
        user_splits=None,
    ):
        ...

    @sarus_method("optbinning.BINNING_TABLE_BUILD")
    def build(
        self,
        show_digits=2,
        add_totals=True,
    ):
        ...


class Scorecard(DataSpecWrapper[optbinning.Scorecard]):
    @sarus_init("optbinning.SCORECARD")
    def __init__(
        self,
        binning_process,
        estimator,
        scaling_method=None,
        scaling_method_params=None,
        intercept_based=False,
        reverse_scorecard=False,
        rounding=False,
        verbose=False,
    ):
        ...

    @sarus_method("optbinning.SCORECARD_FIT", inplace=True)
    def fit(
        self,
        X,
        y,
        sample_weight=None,
        metric_special=0,
        metric_missing=0,
        show_digits=2,
        check_input=False,
    ):
        ...

    @sarus_method("optbinning.SCORECARD_PREDICT")
    def predict(self, X):
        ...

    @sarus_method("optbinning.SCORECARD_PREDICT_PROBA")
    def predict_proba(self, X):
        ...



from __future__ import annotations

import logging
from functools import partial

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init, sarus_method, sarus_property

try:
    from scipy.sparse import (
        bsr_array,
        bsr_matrix,
        coo_array,
        coo_matrix,
        csc_array,
        csc_matrix,
        csr_array,
        csr_matrix,
        dia_array,
        dia_matrix,
        dok_array,
        dok_matrix,
        lil_array,
        lil_matrix,
    )
except ModuleNotFoundError:
    pass

logger = logging.getLogger(__name__)

sarus_method = partial(sarus_method)
sarus_property = partial(sarus_property)
sarus_init = partial(sarus_init)


class bsr_matrix(DataSpecWrapper[bsr_matrix]):
    @sarus_init("scipy.SCIPY_BSR_MATRIX")
    def __init__(
        self,
        arg1,
        shape=None,
        dtype=None,
        copy=False,
        blocksize=None,
    ):
        ...

    @sarus_method("scipy.SCIPY_ELIMINATE_ZEROS", inplace=True)
    def eliminate_zeros(self):
        ...

    @sarus_method("scipy.SCIPY_SORT_INDICES", inplace=True)
    def sorted_indices(self):
        ...

    @sarus_method("scipy.SCIPY_SUM_DUPLICATES", inplace=True)
    def sum_duplicates(self):
        ...


class coo_matrix(DataSpecWrapper[coo_matrix]):
    @sarus_init("scipy.SCIPY_COO_MATRIX")
    def __init__(
        self,
        arg1,
        shape=None,
        dtype=None,
        copy=False,
    ):
        ...

    @sarus_method("scipy.SCIPY_ELIMINATE_ZEROS", inplace=True)
    def eliminate_zeros(self):
        ...

    @sarus_method("scipy.SCIPY_SORT_INDICES", inplace=True)
    def sorted_indices(self):
        ...

    @sarus_method("scipy.SCIPY_SUM_DUPLICATES", inplace=True)
    def sum_duplicates(self):
        ...


class csc_matrix(DataSpecWrapper[csc_matrix]):
    @sarus_init("scipy.SCIPY_CSC_MATRIX")
    def __init__(
        self,
        arg1=None,
        shape=None,
        dtype=None,
        copy=False,
    ):
        ...

    @sarus_method("scipy.SCIPY_ELIMINATE_ZEROS", inplace=True)
    def eliminate_zeros(self):
        ...

    @sarus_method("scipy.SCIPY_SORT_INDICES", inplace=True)
    def sorted_indices(self):
        ...

    @sarus_method("scipy.SCIPY_SUM_DUPLICATES", inplace=True)
    def sum_duplicates(self):
        ...


class csr_matrix(DataSpecWrapper[csr_matrix]):
    @sarus_init("scipy.SCIPY_CSR_MATRIX")
    def __init__(
        self,
        arg1=None,
        shape=None,
        dtype=None,
        copy=False,
    ):
        ...

    @sarus_method("scipy.SCIPY_ELIMINATE_ZEROS", inplace=True)
    def eliminate_zeros(self):
        ...

    @sarus_method("scipy.SCIPY_SORT_INDICES", inplace=True)
    def sorted_indices(self):
        ...

    @sarus_method("scipy.SCIPY_SUM_DUPLICATES", inplace=True)
    def sum_duplicates(self):
        ...


class dia_matrix(DataSpecWrapper[dia_matrix]):
    @sarus_init("scipy.SCIPY_DIA_MATRIX")
    def __init__(
        self,
        arg1,
        shape=None,
        dtype=None,
        copy=False,
    ):
        ...

    @sarus_method("scipy.SCIPY_ELIMINATE_ZEROS", inplace=True)
    def eliminate_zeros(self):
        ...

    @sarus_method("scipy.SCIPY_SORT_INDICES", inplace=True)
    def sorted_indices(self):
        ...

    @sarus_method("scipy.SCIPY_SUM_DUPLICATES", inplace=True)
    def sum_duplicates(self):
        ...


class dok_matrix(DataSpecWrapper[dok_matrix]):
    @sarus_init("scipy.SCIPY_DOK_MATRIX")
    def __init__(
        self,
        arg1=None,
        shape=None,
        dtype=None,
        copy=False,
    ):
        ...

    @sarus_method("scipy.SCIPY_ELIMINATE_ZEROS", inplace=True)
    def eliminate_zeros(self):
        ...

    @sarus_method("scipy.SCIPY_SORT_INDICES", inplace=True)
    def sorted_indices(self):
        ...

    @sarus_method("scipy.SCIPY_SUM_DUPLICATES", inplace=True)
    def sum_duplicates(self):
        ...


class lil_matrix(DataSpecWrapper[lil_matrix]):
    @sarus_init("scipy.SCIPY_LIL_MATRIX")
    def __init__(
        self,
        arg1=None,
        shape=None,
        dtype=None,
        copy=False,
    ):
        ...

    @sarus_method("scipy.SCIPY_ELIMINATE_ZEROS", inplace=True)
    def eliminate_zeros(self):
        ...

    @sarus_method("scipy.SCIPY_SORT_INDICES", inplace=True)
    def sorted_indices(self):
        ...

    @sarus_method("scipy.SCIPY_SUM_DUPLICATES", inplace=True)
    def sum_duplicates(self):
        ...


class bsr_array(DataSpecWrapper[bsr_array]):
    @sarus_init("scipy.SCIPY_BSR_ARRAY")
    def __init__(
        self,
        arg1,
        shape=None,
        dtype=None,
        copy=False,
        blocksize=None,
    ):
        ...

    @sarus_method("scipy.SCIPY_ELIMINATE_ZEROS", inplace=True)
    def eliminate_zeros(self):
        ...

    @sarus_method("scipy.SCIPY_SORT_INDICES", inplace=True)
    def sorted_indices(self):
        ...

    @sarus_method("scipy.SCIPY_SUM_DUPLICATES", inplace=True)
    def sum_duplicates(self):
        ...


class coo_array(DataSpecWrapper[coo_array]):
    @sarus_init("scipy.SCIPY_COO_ARRAY")
    def __init__(self, arg1, shape=None, dtype=None, copy=False):
        ...

    @sarus_method("scipy.SCIPY_ELIMINATE_ZEROS", inplace=True)
    def eliminate_zeros(self):
        ...

    @sarus_method("scipy.SCIPY_SORT_INDICES", inplace=True)
    def sorted_indices(self):
        ...

    @sarus_method("scipy.SCIPY_SUM_DUPLICATES", inplace=True)
    def sum_duplicates(self):
        ...


class csc_array(DataSpecWrapper[csc_array]):
    @sarus_init("scipy.SCIPY_CSC_ARRAY")
    def __init__(self, arg1, shape=None, dtype=None, copy=False):
        ...

    @sarus_method("scipy.SCIPY_ELIMINATE_ZEROS", inplace=True)
    def eliminate_zeros(self):
        ...

    @sarus_method("scipy.SCIPY_SORT_INDICES", inplace=True)
    def sorted_indices(self):
        ...

    @sarus_method("scipy.SCIPY_SUM_DUPLICATES", inplace=True)
    def sum_duplicates(self):
        ...


class csr_array(DataSpecWrapper[csr_array]):
    @sarus_init("scipy.SCIPY_CSR_ARRAY")
    def __init__(self, arg1, shape=None, dtype=None, copy=False):
        ...

    @sarus_method("scipy.SCIPY_ELIMINATE_ZEROS", inplace=True)
    def eliminate_zeros(self):
        ...

    @sarus_method("scipy.SCIPY_SORT_INDICES", inplace=True)
    def sorted_indices(self):
        ...

    @sarus_method("scipy.SCIPY_SUM_DUPLICATES", inplace=True)
    def sum_duplicates(self):
        ...


class dia_array(DataSpecWrapper[dia_array]):
    @sarus_init("scipy.SCIPY_DIA_ARRAY")
    def __init__(self, arg1, shape=None, dtype=None, copy=False):
        ...

    @sarus_method("scipy.SCIPY_ELIMINATE_ZEROS", inplace=True)
    def eliminate_zeros(self):
        ...

    @sarus_method("scipy.SCIPY_SORT_INDICES", inplace=True)
    def sorted_indices(self):
        ...

    @sarus_method("scipy.SCIPY_SUM_DUPLICATES", inplace=True)
    def sum_duplicates(self):
        ...


class dok_array(DataSpecWrapper[dok_array]):
    @sarus_init("scipy.SCIPY_DOK_ARRAY")
    def __init__(self, arg1, shape=None, dtype=None, copy=False):
        ...

    @sarus_method("scipy.SCIPY_ELIMINATE_ZEROS", inplace=True)
    def eliminate_zeros(self):
        ...

    @sarus_method("scipy.SCIPY_SORT_INDICES", inplace=True)
    def sorted_indices(self):
        ...

    @sarus_method("scipy.SCIPY_SUM_DUPLICATES", inplace=True)
    def sum_duplicates(self):
        ...


class lil_array(DataSpecWrapper[lil_array]):
    @sarus_init("scipy.SCIPY_LIL_ARRAY")
    def __init__(self, arg1, shape=None, dtype=None, copy=False):
        ...

    @sarus_method("scipy.SCIPY_ELIMINATE_ZEROS", inplace=True)
    def eliminate_zeros(self):
        ...

    @sarus_method("scipy.SCIPY_SORT_INDICES", inplace=True)
    def sorted_indices(self):
        ...

    @sarus_method("scipy.SCIPY_SUM_DUPLICATES", inplace=True)
    def sum_duplicates(self):
        ...


register_ops()



from __future__ import annotations

import inspect
import logging
import re
import traceback
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
    get_args,
)

import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st
from sarus_data_spec.context import global_context
from sarus_data_spec.dataspec_validator.typing import DataspecPrivacyPolicy
from sarus_data_spec.status import DataSpecErrorStatus
from sarus_data_spec.transform import variable

import sarus.manager.dataspec_api as api
from sarus.context.typing import LocalSDKContext
from sarus.manager.typing import SDKManager
from sarus.typing import DataSpecVariant, SyncPolicy
from sarus.utils import register_ops

try:
    from IPython.core.display import HTML
except ModuleNotFoundError:
    pass

logger = logging.getLogger(__name__)


IGNORE_WARNING = [
    "_ipython_canary_method_should_not_exist_",
    "_repr_png_",
]

T = TypeVar("T")


def class_name(_class: type) -> str:
    return getattr(_class, "__name__", str(_class))


class MetaWrapper(type):
    """Metaclass to delegate some class attributes to the wrapped class."""

    _wrapper_classes: List[Tuple[str, type]] = []

    def __new__(cls, name, bases, dct):
        """Creation of a new DataSpecWrapper class.

        The arguments `name`, `bases` and `dct` are the standard arguments
        passed to the `type` function when creating a new class
        https://docs.python.org/3/library/functions.html#type
        """
        new_wrapper_class = super().__new__(cls, name, bases, dct)

        if name != "DataSpecWrapper":
            # Set the wrapped class from the type annotation
            wrapped_python_class = get_args(
                new_wrapper_class.__orig_bases__[0]
            )[0]
            new_wrapper_class.__wraps__ = wrapped_python_class

            # Register the new class in the wrapper factory
            python_classname = str(wrapped_python_class)
            if python_classname == "~T":
                logger.warning(
                    f"Wrapper class {name} has no associated Python type."
                )
            else:
                MetaWrapper._wrapper_classes.append(
                    (str(wrapped_python_class), new_wrapper_class)
                )

        return new_wrapper_class

    def __getattr__(self, name):
        return getattr(self.__wraps__, name)


class DataSpecWrapper(Generic[T], metaclass=MetaWrapper):
    """This class wraps Sarus DataSpecs objects for the SDK.

    More specifically, it wraps 3 variants of the same DataSpec. These variants
    can be identified with the DataSpecVariant enum:
    - USER_DEFINED: the DataSpec as defined by the user
    - SYNTHETIC: the synthetic variant of the USER_DEFINED DataSpec
    - MOCK: a small sample of the SYNTHETIC variant

    The wrapper has a fallback behavior implemented in the __getattr__ method.
    Any attribute access attempt that is not explicitly catched by a method
    will be delegated to the SYNTHETIC dataspec's value.

    Subclasses such as sarus.pandas.DataFrame must implement the `value` method.
    """

    # This is a class variable hoding all DataSpecWrapper instances
    # This is used by the `dataspec_wrapper_symbols` method.
    instances: Dict[str, Tuple[int, str]] = dict()

    @classmethod
    def from_dataspec(
        cls, dataspec: st.DataSpec
    ) -> DataSpecWrapper(Generic[T]):
        """Instantiate a DataspecWrapper from a dataspec.

        We don't use the __init__ method because the __new__ method may be
        overwritten by the `sarus.utils.init_wrapped` decorator.
        """
        wrapper = object.__new__(cls)
        wrapper._set_dataspec(dataspec)
        return wrapper

    @classmethod
    def from_value(cls, value: T) -> DataSpecWrapper(Generic[T]):
        """Construct a DataSpecWrapper by wrapping a Python value."""
        return cls(value)

    def _set_dataspec(self, dataspec: st.DataSpec) -> None:
        self._dataspec = dataspec
        self._alt_dataspec = None
        self._alt_policy = None  # The alternative Dataspec's privacy policy
        self._traced_transform = None

        # This class only works with a LocalSDKContext
        context: LocalSDKContext = global_context()
        assert isinstance(context, LocalSDKContext)
        self._manager = context.manager()
        if context.sync_policy() == SyncPolicy.SEND_ON_INIT:
            self.alternative_dataspec(launch=True)
        # for .dot() to work add a else condition with self.alternative_dataspec()

        # Register wrapper instance
        DataSpecWrapper.instances[dataspec.uuid()] = id(self), class_name(
            self.__wraps__
        )

    def _set_variable(self, name: str, position: int):
        """Used to trace callables."""
        self._traced_transform = variable(name=name, position=position)

    def traced_transform(self) -> Optional[st.Transform]:
        """Return the wrapper's composed transform or variable when tracing."""
        return self._traced_transform

    def _set_traced_transform(
        self, traced_transform: Optional[st.Transform]
    ) -> None:
        self._traced_transform = traced_transform

    def manager(self) -> SDKManager:
        return self._manager

    def python_type(self) -> Optional[str]:
        """Return the value's Python type name.

        The SDKManager registers an attribute holding the MOCK value's
        Python type (see `SDKManager.infer_output_type` method). This is
        used to instantiate the right DataSpecWrapper class e.g. instantiate a
        sarus.pandas.DataFrame if the Python value is a pandas.DataFrame.
        """
        return self.manager().python_type(self._dataspec)

    def dataspec_wrapper_symbols(
        self, f_locals, f_globals
    ) -> Dict[str, Optional[str]]:
        """Returns the symbols table in the caller's namespace.

        For instance, if the data practitioner defines a DataSpecWrapper using
        the symbol X in his code (e.g. X = dataset.as_pandas()) then the symbol
        table contain the mapping between the DataSpecWrapper instances' ids
        and their symbols. This is used to make the dot representation more
        readable.
        """

        def is_ipython_variable(name: str) -> bool:
            return bool(re.search(r"^_\d+$", name))

        mapping = {
            id(obj): (name, class_name(obj.__wraps__))
            for name, obj in f_locals.items()
            if isinstance(obj, DataSpecWrapper)
            and not is_ipython_variable(name)
        }
        mapping.update(
            {
                id(obj): (name, class_name(obj.__wraps__))
                for name, obj in f_globals.items()
                if isinstance(obj, DataSpecWrapper)
                and not is_ipython_variable(name)
            }
        )
        names = dict()
        storage = self.manager().storage()
        rewriter = self.manager().dataspec_rewriter()
        for uuid, (_id, _class) in DataSpecWrapper.instances.items():
            ds = storage.referrable(uuid)
            info = mapping[_id] if _id in mapping else ("", _class)
            names[uuid] = info
            # The symbol & class is also valid for all variants
            for variant in rewriter.variants(ds):
                names[variant.uuid()] = info
        return names

    def dot(
        self,
        kind: Union[
            DataSpecVariant,
            Literal["rewritten", "mock", "original", "synthetic"],
        ] = "rewritten",
        remote: bool = True,
    ) -> str:
        """Graphviz's dot representation of the DataSpecWrapper graph.

        Uses color codes to show statuses.

        Args:
            kind (DataSpecVariant): the DataSpec to represent.
            remote (true): Which Manager to inspect.
                If true shows the DataSpec's status on the server,
                else show the DataSpec's status locally.
        """
        if isinstance(kind, str):
            kind = DataSpecVariant(kind)
        _, policy = self.alternative_dataspec()  # send to the server
        ds = self.dataspec(kind)
        if kind != DataSpecVariant.ALTERNATIVE:
            policy = None

        # pull attributes linking the variants
        api.pull_dataspec_graph(self.manager().client(), self._dataspec.uuid())

        caller_frame = inspect.currentframe().f_back
        symbols = self.dataspec_wrapper_symbols(
            caller_frame.f_locals, caller_frame.f_globals
        )
        return self.manager().dot(
            ds, symbols=symbols, remote=remote, policy=policy
        )

    def alternative_dataspec(
        self,
        target_epsilon: Optional[float] = None,
        launch: bool = False,
    ) -> Tuple[st.DataSpec, DataspecPrivacyPolicy]:
        """Send the DataSpec to the server for rewriting.

        The server sends an alternative DataSpec back which is compliant with
        the privacy policy defined for the current user.

        Args:
            launch (bool): If true, tell the server to launch the computation.
            target_epsilon (float, optional): The target_epsilon to evaluate
                the dataspec.

        Returns:
            A tuple with the alternative dataspec and the corresponding privacy
            policy.
        """
        context: LocalSDKContext = global_context()
        dataspec = self.dataspec(kind=DataSpecVariant.USER_DEFINED)

        if context.sync_policy() == SyncPolicy.MANUAL:
            # In MANUAL mode we evaluate on Synthetic data
            syn_ds = dataspec.variant(st.ConstraintKind.SYNTHETIC)
            return syn_ds, DataspecPrivacyPolicy.SYNTHETIC

        # Ask for a new rewriting
        if self._alt_dataspec is None or target_epsilon is not None:
            self.manager().push(dataspec)
            alt_dataspec, alt_policy = self.manager().rewrite(
                dataspec, target_epsilon
            )
            # Register the new dataspec as being represented by this wrapper
            DataSpecWrapper.instances[alt_dataspec.uuid()] = (
                id(self),
                class_name(self.__wraps__),
            )

            # Remember value for default epsilon
            if target_epsilon is None:
                self._alt_dataspec = alt_dataspec
                self._alt_policy = alt_policy
        else:
            # return already computed default alternative
            alt_dataspec = self._alt_dataspec
            alt_policy = self._alt_policy

        if launch:
            computation = self.manager().dataspec_computation(alt_dataspec)
            if self.manager().status(dataspec, computation.task_name) is None:
                self.manager().launch(alt_dataspec)

        return alt_dataspec, alt_policy

    def delete_from_local(self) -> None:
        """Delete a DataSpec from the local storage."""
        dataspec = self.dataspec(kind=DataSpecVariant.ALTERNATIVE)
        self.manager()._delete_local(dataspec.uuid())

    def __eval_policy__(self) -> str:
        """The alternative dataspec's privacy policy."""
        _, alt_policy = self.alternative_dataspec()
        return alt_policy.value

    def _status(self) -> Dict[str, Any]:
        dataspecs = {
            "user": self.dataspec(kind=DataSpecVariant.USER_DEFINED),
            "alt": self.dataspec(kind=DataSpecVariant.ALTERNATIVE),
            "mock": self.dataspec(kind=DataSpecVariant.MOCK),
        }

        result = {}
        for key, ds in dataspecs.items():
            statuses = self._manager._get_status(ds, remote=True)
            remote_status = [s for s in statuses if s["id"] == ds.uuid()][0]

            statuses = self._manager._get_status(ds, remote=False)
            local_status = [s for s in statuses if s["id"] == ds.uuid()][0]

            result[key] = {
                "remote": remote_status.get("code", "no_status"),
                "local": local_status.get("code", "no_status"),
            }

        return result

    def dataspec(
        self, kind: DataSpecVariant = DataSpecVariant.USER_DEFINED
    ) -> st.DataSpec:
        """Return one of the wrapped DataSpec object."""
        if kind == DataSpecVariant.USER_DEFINED:
            return self._dataspec
        if kind == DataSpecVariant.ALTERNATIVE:
            if self._alt_dataspec:
                return self._alt_dataspec
            else:
                return self._dataspec.variant(kind=st.ConstraintKind.SYNTHETIC)
        elif kind == DataSpecVariant.SYNTHETIC:
            return self._dataspec.variant(kind=st.ConstraintKind.SYNTHETIC)
        elif kind == DataSpecVariant.MOCK:
            return self._dataspec.variant(kind=st.ConstraintKind.MOCK)
        else:
            raise ValueError(f"Unknown kind {kind}")

    def save(self, path: str) -> None:
        """Save the representation of the sarus object to a file.

        Args:
            path (str): The path to save the object to.
        """
        self._manager._client.save(self, path)

    def __len__(self) -> int:
        logger.info(
            f"`__len__` not supported on {type(self)}, "
            "object has been evaluated for this method. "
            "See Sarus documentation."
        )
        return self.__sarus_eval__().__len__()

    def __repr__(self) -> int:
        return self.__sarus_eval__().__repr__()

    def __iter__(self) -> int:
        return self.__sarus_eval__().__iter__()

    def __float__(self) -> int:
        return self.__sarus_eval__().__float__()

    def __int__(self) -> int:
        return self.__sarus_eval__().__int__()

    def __bool__(self) -> int:
        return self.__sarus_eval__().__bool__()

    def __format__(self, *args, **kwargs) -> int:
        return self.__sarus_eval__().__format__(*args, **kwargs)

    def __sarus_eval__(
        self,
        target_epsilon: Optional[float] = None,
        verbose: Optional[int] = None,
    ) -> T:
        """Return the value of alternative DataSpec's variant."""
        alt_dataspec, alt_policy = self.alternative_dataspec(target_epsilon)
        if verbose is None:
            context: LocalSDKContext = global_context()
            verbose = context.verbose()

        if (
            target_epsilon is not None
            and target_epsilon > 0
            and alt_policy == DataspecPrivacyPolicy.SYNTHETIC
            and verbose >= 1
        ):
            logger.warning(
                "Warning: target_epsilon could not be consumed. "
                "This might be due to privacy budget "
                "constraints or some operations not "
                "rewritable with Differential Privacy."
            )

        if verbose >= 1:
            (
                has_transform,
                transform_to_small_data,
            ) = self.manager().has_transform(alt_dataspec, 'ToSmallData')
            if has_transform:
                to_small_data_size = (
                    transform_to_small_data.protobuf().spec.to_small_data.size
                )
                logger.warning(
                    "Due to the large size of the dataset, "
                    "the transforms were applied to a sample of the "
                    f"data only (sample_size={to_small_data_size}), "
                    "so your results may be affected.",
                    stacklevel=0,
                )

            message = alt_policy.value
            final_message = f"{message} (uuid:{self._dataspec.uuid()}, alt_uuid:{alt_dataspec.uuid()})"

            # check public result come from public table
            if alt_policy == DataspecPrivacyPolicy.PUBLIC:
                if any(
                    not source.is_transformed()
                    for source in alt_dataspec.sources()
                ):
                    print(final_message)
            else:
                if alt_policy == DataspecPrivacyPolicy.DP:
                    epsilon = self.manager().consumed_epsilon(alt_dataspec)
                    final_message = f"{message} (epsilon={epsilon:.2f}, uuid:{self._dataspec.uuid()}, alt_uuid:{alt_dataspec.uuid()})"

                print(final_message)

        if alt_dataspec.prototype() == sp.Dataset:
            dataset = cast(st.Dataset, alt_dataspec)
            value = dataset.to(self.__wraps__)
        else:
            scalar = cast(st.Scalar, alt_dataspec)
            value = cast(st.Scalar, scalar.value())

        return value

    def __getattr__(self, name: str) -> Any:
        """Called when trying to access an attribute of the DataSpecWrapper or calling a non-bound method.

        If this method is called, we have already passed by the default Python
        self.__getattribute__(name) and it found nothing.
        """
        if name not in IGNORE_WARNING:
            logger.info(
                f"`{name}` not supported on {type(self)}, "
                "object has been evaluated for this method. "
                "See Sarus documentation."
            )
            if self.traced_transform() is not None:
                raise NotImplementedError(
                    f"Cannot trace `{name}`. "
                    f"Not supported on type {self.__wraps__}"
                )
            return self.__sarus_eval__().__getattribute__(name)
        else:
            # IPython expects an error to be raised on some attributes.
            return object.__getattribute__(
                self, name
            )  # used to raise an error directly

    def _ipython_display_(self) -> None:
        try:
            display(self.__sarus_eval__())  # noqa: F821
        except Exception as e:
            if isinstance(e, DataSpecErrorStatus):
                display(e.extract_last_exc() or str(e))  # noqa: F821
            else:
                display(f"Exception: {e}")  # noqa: F821

            tb = traceback.format_exc()
            html_traceback = f"""
                <details>
                    <summary style="color: red;">Click to view traceback</summary>
                    <pre>{tb}</pre>
                </details>
                """
            display(HTML(html_traceback))  # noqa: F821

    def _mock_value(self) -> Any:
        """Debug purposes."""
        mock_ds = self.dataspec(DataSpecVariant.MOCK)
        if mock_ds.prototype() == sp.Dataset:
            return mock_ds.to_pandas()
        else:
            return mock_ds.value()


register_ops(use_parent_module=True)



from __future__ import annotations

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import sarus_init


class Slice(DataSpecWrapper[slice]):
    @sarus_init("std.SLICE")
    def __init__(self, *args, **kwargs) -> None:
        ...

    @classmethod
    def from_value(cls, value: slice) -> Slice:
        elems = [
            value.start,
            value.stop,
            value.step,
        ]
        return Slice(*elems)


class Set(DataSpecWrapper[set]):
    @sarus_init("std.SET")
    def __init__(self, *args, **kwargs) -> None:
        """Need to pass all arguments Set(*args)."""
        ...

    @classmethod
    def from_value(cls, value: set) -> Set:
        elems = [e for e in value]
        return Set(*elems)


class Dict(DataSpecWrapper[dict]):
    @sarus_init("std.DICT")
    def __init__(self, *args, **kwargs) -> None:
        """Need to pass all arguments Dict(**kwargs)."""
        ...

    @classmethod
    def from_value(cls, value: dict) -> Dict:
        elems = {k: v for k, v in value.items()}
        return Dict(**elems)


class List(DataSpecWrapper[list]):
    @sarus_init("std.LIST")
    def __init__(self, *args, **kwargs) -> None:
        """Need to pass all arguments List(*args)."""

    def __iter__(self):
        """The iterator protocol is emulated using __getitem__.

        X_train, X_test, y_train, y_test = train_test_split(...)

        Under the hood it does:
        result: DataSpecWrapper[List] = train_test_split(...)
        X_train = result.__getitem__[0]
        X_test = result.__getitem__[1]
        y_train = result.__getitem__[2]
        y_test = result.__getitem__[3]
        """
        self.__sarus_idx__ = 0
        return self

    def __next__(self):
        try:
            idx = self.__sarus_idx__
            self.__sarus_idx__ += 1
            return self[idx]
        except IndexError:
            raise StopIteration

    @classmethod
    def from_value(cls, value: list) -> List:
        elems = [e for e in value]
        return List(*elems)


class Tuple(DataSpecWrapper[tuple]):
    @sarus_init("std.TUPLE")
    def __init__(self, *args, **kwargs) -> None:
        """Need to pass all arguments Tuple(*args)."""

    # The iterator protocol is emulated using __getitem__
    def __iter__(self):
        self.__sarus_idx__ = 0
        return self

    def __next__(self):
        try:
            idx = self.__sarus_idx__
            self.__sarus_idx__ += 1
            return self[idx]
        except IndexError:
            raise StopIteration

    @classmethod
    def from_value(cls, value: tuple) -> Tuple:
        elems = [e for e in value]
        return Tuple(*elems)


class Int(DataSpecWrapper[int]):
    @sarus_init("std.INT")
    def __init__(self, *args, **kwargs) -> None:
        ...


class Float(DataSpecWrapper[float]):
    @sarus_init("std.FLOAT")
    def __init__(self, *args, **kwargs) -> None:
        ...


class String(DataSpecWrapper[str]):
    @sarus_init("std.STRING")
    def __init__(self, *args, **kwargs) -> None:
        ...


class Bool(DataSpecWrapper[bool]):
    @sarus_init("std.BOOL")
    def __init__(self, *args, **kwargs) -> None:
        ...



import numpy as np

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init, sarus_method

try:
    import sklearn.preprocessing as sk_preprocessing
    from sklearn.preprocessing import *  # noqa: F401
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class OneHotEncoder(DataSpecWrapper[sk_preprocessing.OneHotEncoder]):
    @sarus_init("sklearn.SK_ONEHOT")
    def __init__(
        self,
        *,
        categories="auto",
        drop=None,
        sparse=True,
        dtype=np.float64,
        handle_unknown="error",
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_PIPELINE_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


class LabelEncoder(DataSpecWrapper[sk_preprocessing.LabelEncoder]):
    @sarus_init("sklearn.SK_LABEL_ENCODER")
    def __init__(
        self,
        *,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT_Y", inplace=True)
    def fit(self, y):
        ...


class FunctionTransformer(
    DataSpecWrapper[sk_preprocessing.FunctionTransformer]
):
    @sarus_init("sklearn.SK_FUNCTION_TRANSFORMER")
    def __init__(
        self,
        func=None,
        inverse_func=None,
        *,
        validate=False,
        accept_sparse=False,
        check_inverse=True,
        feature_names_out=None,
        kw_args=None,
        inv_kw_args=None,
    ):
        ...

    @sarus_method("sklearn.SK_PIPELINE_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


register_ops()



import pandas as pd

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_property


class Index(DataSpecWrapper[pd.Index]):
    ...


class RangeIndex(DataSpecWrapper[pd.RangeIndex]):
    ...


class Int64Index(DataSpecWrapper[pd.Int64Index]):
    ...


class UInt64Index(DataSpecWrapper[pd.UInt64Index]):
    ...


class Float64Index(DataSpecWrapper[pd.Float64Index]):
    ...


class DataFrameGroupBy(DataSpecWrapper[pd.core.groupby.DataFrameGroupBy]):
    @sarus_property("pandas.PD_GROUPS_GROUPBY")
    def groups(self):
        ...


class SeriesGroupBy(DataSpecWrapper[pd.core.groupby.SeriesGroupBy]):
    @sarus_property("pandas.PD_GROUPS_GROUPBY")
    def groups(self):
        ...


class RollingGroupby(DataSpecWrapper[pd.core.window.rolling.RollingGroupby]):
    ...


register_ops()
